import bionumpy as bnp                                                                                                                                                                                                                 
import time


t = time.perf_counter()
f = bnp.open("test.fa", buffer_type=bnp.TwoLineFastaBuffer)
for chunk in f.read_chunks(min_chunk_size=100000000):
    print(time.perf_counter()-t)


t = time.perf_counter()
raw_chunks = bnp.io.parser.NumpyFileReader(open("test.fa", "rb"), bnp.TwoLineFastaBuffer).read_chunks(min_chunk_size=100000000)
for chunk in raw_chunks:
    print(chunk)
    print(time.perf_counter()-t)
    print(chunk.get_data().sequence)
    print(time.perf_counter()-t)



#cython: language_level=3
from libc.stdio cimport *
import numpy as np
cimport numpy as np
import logging
import time
cimport cython
from graph_kmer_index.shared_mem import to_shared_memory, SingleSharedArray
from kmer_mapper.util import log_memory_usage_now
import gc
from libc.stdlib cimport free
from cython.parallel import prange


@cython.nonecheck(False)
@cython.cdivision(True)
@cython.boundscheck(False)
@cython.wraparound(False)
def map_kmers_to_graph_index(index, int max_node_id, np.uint64_t[::1] kmers, int max_index_lookup_frequency=1000):
    t = time.perf_counter()
    # index arrays
    cdef np.int32_t[::1] hashes_to_index = index._hashes_to_index
    #cdef np.ndarray[np.int32_t] hashes_to_index = index._hashes_to_index
    cdef np.int32_t[::1] n_kmers = index._n_kmers
    #cdef np.ndarray[np.uint32_t] n_kmers = index._n_kmers
    cdef np.int32_t[::1] nodes = index._nodes
    cdef np.uint64_t[::1] index_kmers = index._kmers
    cdef np.uint16_t[::1] index_frequencies = index._frequencies
    cdef unsigned long modulo = index._modulo

    cdef int n_local_hits
    cdef int index_position

    cdef int i = 0
    cdef int l = 0
    cdef int j = 0
    cdef np.ndarray[np.uint32_t] node_counts = np.zeros(max_node_id+1, dtype=np.uint32)
    cdef np.uint32_t[::1] node_counts_view = node_counts


    t = time.perf_counter()
    #cdef np.uint64_t[::1] kmer_hashes = np.mod(kmers, modulo)
    cdef int n_collisions = 0
    cdef int n_kmers_mapped = 0
    cdef int n_skipped_high_frequency = 0
    cdef int n_no_index_hits = 0
    cdef unsigned long kmerhash
    t = time.perf_counter()

    #cdef np.uint32_t[::1] n_local_hits_array = n_kmers[kmer_hashes]
    #cdef np.int64_t[::1] index_position_array = hashes_to_index[kmer_hashes]

    for i in range(kmers.shape[0]):
        kmerhash = kmers[i] % modulo  # kmer_hashes[i]
        n_local_hits = n_kmers[kmerhash]
        index_position = hashes_to_index[kmerhash]
        l = index_position
        for j in range(n_local_hits):
            # Check that this entry actually matches the kmer, sometimes it will not due to collision
            if index_kmers[l] != kmers[i]:
                l += 1
                continue

            if index_frequencies[l] > max_index_lookup_frequency:
                l += 1
                continue

            node_counts[nodes[l]] += 1
            l += 1

    logging.debug("Time spent looking up kmers in index: %.3f" % (time.perf_counter()-t))
    return node_counts



@cython.nonecheck(False)
@cython.cdivision(True)
@cython.boundscheck(False)
@cython.wraparound(False)
# returns an array of same size of kmers with True for kmers that are in index
def in_graph_index(index,  np.uint64_t[::1] kmers, int max_index_lookup_frequency=1000):
    t = time.perf_counter()
    # index arrays
    cdef np.int32_t[::1] hashes_to_index = index._hashes_to_index
    #cdef np.ndarray[np.int32_t] hashes_to_index = index._hashes_to_index
    cdef np.int32_t[::1] n_kmers = index._n_kmers
    #cdef np.ndarray[np.uint32_t] n_kmers = index._n_kmers
    cdef np.int32_t[::1] nodes = index._nodes
    cdef np.uint64_t[::1] index_kmers = index._kmers
    cdef np.uint16_t[::1] index_frequencies = index._frequencies
    cdef unsigned long modulo = index._modulo

    cdef int n_local_hits
    cdef int index_position

    cdef long i = 0
    cdef long l = 0
    cdef long j = 0
    cdef np.ndarray[np.uint8_t] out= np.zeros(len(kmers), dtype=np.uint8)

    t = time.perf_counter()
    cdef int n_collisions = 0
    cdef int n_kmers_mapped = 0
    cdef int n_skipped_high_frequency = 0
    cdef int n_no_index_hits = 0
    cdef unsigned long kmerhash
    t = time.perf_counter()


    cdef unsigned short hit = False

    for i in range(kmers.shape[0]):
    #for i in prange(0, kmers.shape[0], num_threads=8, nogil=True):
        hit = 0
        kmerhash = kmers[i] % modulo  # kmer_hashes[i]
        n_local_hits = n_kmers[kmerhash]
        index_position = hashes_to_index[kmerhash]
        l = index_position
        for j in range(n_local_hits):
            # Check that this entry actually matches the kmer, sometimes it will not due to collision
            if index_kmers[l] != kmers[i]:
                l += 1
                continue
            hit = True
            break

        out[i] = hit

    logging.debug("Time spent looking up kmers in index: %.3f" % (time.perf_counter()-t))
    return out



import time
import numpy as np
from kmer_mapper.gpu_counter import GpuCounter
import cupy as cp
import bionumpy as bnp
import pytest


@pytest.mark.skip()
def test_bionumpy_numpy():
    file = bnp.open("test.fa", buffer_type=bnp.TwoLineFastaBuffer)
    for i, chunk in enumerate(file.read_chunks(min_chunk_size=5000000)):
        if i > 10:
            break

    return


@pytest.mark.skip()
def test_bionumpy_cupy():
    bnp.set_backend(cp)
    file = bnp.open("test.fa", buffer_type=bnp.TwoLineFastaBuffer)
    for i, chunk in enumerate(file.read_chunks(min_chunk_size=5000000)):
        if i > 10:
            break

    return


@pytest.mark.skip()
def test_time():
    funcs = [test_bionumpy_numpy, test_bionumpy_cupy]
    for func in funcs:
        t0 = time.perf_counter()
        func()
        print(func, time.perf_counter()-t0)


@pytest.mark.skip()
def test_count():
    kmers = np.array([1, 2, 3])
    nodes = np.array([10, 11, 12])

    counter = GpuCounter(kmers, kmers, nodes)
    counter.initialize_cuda(2003)
    counter.count(np.array([1, 1, 1, 2, 3, 1, 3]))
    node_counts = counter.get_node_counts(15)
    assert np.all(node_counts[[10, 11, 12]] == [4, 1, 2])
    print("Passed")


if __name__ == "__main__":
    test_count()
    test_time()


from itertools import product
import numpy as np

class BaseEncoding:
    """ Basic ACII byte encoding """

    @classmethod
    def from_string(cls, sequence):
        return np.array([ord(c) for c in sequence], dtype=np.uint8)

    @classmethod
    def from_bytes(cls, sequence):
        """Identity"""
        return sequence

    @classmethod
    def to_bytes(cls, sequence):
        """Identity"""
        return sequence

    @classmethod
    def to_string(cls, byte_sequence):
        return "".join(chr(b) for b in byte_sequence)

class ACTGTwoBitEncoding:
    letters = ["A", "C", "T", "G"]
    bitcodes = ["00", "01", 
                "10", "11"]
    reverse = np.array([1, 3, 20, 7], dtype=np.uint8)
    _lookup_2bytes_to_4bits = np.zeros(256*256, dtype=np.uint8)
    _lookup_2bytes_to_4bits[256*reverse[np.arange(4)[:, None]]+reverse[np.arange(4)]] = np.arange(4)[:, None]*4+np.arange(4)
    _shift_4bits = (4*np.arange(2, dtype=np.uint8))
    _shift_2bits = 2*np.arange(4, dtype=np.uint8)

    @classmethod
    def convert_2bytes_to_4bits(cls, two_bytes):
        assert two_bytes.dtype == np.uint16, two_bytes.dtype
        return cls._lookup_2bytes_to_4bits[two_bytes]

    @classmethod
    def join_4bits_to_byte(cls, four_bits):
        return np.bitwise_or.reduce(four_bits << cls._shift_4bits, axis=1)

    @classmethod
    def complement(cls, char):
        complements = np.packbits([1, 0, 1, 0, 1, 0, 1, 0])
        dtype = char.dtype
        return (char.view(np.uint8) ^ complements).view(dtype)

    @classmethod
    def from_bytes(cls, sequence):
        assert sequence.dtype==np.uint8
        assert sequence.size % 4 == 0, sequence.size
        sequence = sequence & 31
        four_bits = cls.convert_2bytes_to_4bits(sequence.view(np.uint16))
        codes = cls.join_4bits_to_byte(four_bits.reshape(-1, 2))
        assert codes.dtype == np.uint8, codes.dtype
        return codes.flatten().view(np.uint8)

    @classmethod
    def from_string(cls, string):
        byte_repr = np.array([ord(c) for c in string], dtype=np.uint8)
        return cls.from_bytes(byte_repr)

    @classmethod
    def to_string(cls, bits):
        byte_repr = cls.to_bytes(bits)
        return "".join(chr(b) for b in byte_repr)

    @classmethod
    def to_bytes(cls, sequence):
        assert sequence.dtype==np.uint8
        bit_mask = np.uint8(3) # last two bits
        all_bytes = (sequence[:, None]>>cls._shift_2bits) & bit_mask
        return cls.reverse[all_bytes.flatten()]+96


class SimpleEncoding(ACTGTwoBitEncoding):
    _lookup_byte_to_2bits = np.zeros(256, dtype=np.uint8)
    _lookup_byte_to_2bits[[97, 65]] = 0
    _lookup_byte_to_2bits[[99, 67]] = 1
    _lookup_byte_to_2bits[[116, 84]] = 2
    _lookup_byte_to_2bits[[103, 71]] = 3

    _shift_2bits = 2*np.arange(4, dtype=np.uint8)

    @classmethod
    def convert_byte_to_2bits(cls, one_byte):
        assert one_byte.dtype == np.uint8, one_byte.dtype
        return cls._lookup_byte_to_2bits[one_byte]

    @classmethod
    def join_2bits_to_byte(cls, two_bits_vector):
        return np.bitwise_or.reduce(two_bits_vector << cls._shift_2bits, axis=-1)

    @classmethod
    def from_bytes(cls, sequence):
        assert sequence.dtype==np.uint8
        assert sequence.size % 4 == 0, sequence.size
        two_bits = cls.convert_byte_to_2bits(sequence)
        codes = cls.join_2bits_to_byte(two_bits.reshape(-1, 4))
        return codes.flatten()

def twobit_swap(number):
    dtype = number.dtype
    byte_lookup = np.zeros(256, dtype=np.uint8)
    power_array = 4**np.arange(4)
    rev_power_array = power_array[::-1]
    for two_bit_string in product([0, 1, 2, 3], repeat=4):
        byte_lookup[np.sum(power_array*two_bit_string)] = np.sum(rev_power_array*two_bit_string)
    new_bytes = byte_lookup[number.view(np.uint8)]
    return new_bytes.view(dtype).byteswap()



import gzip
import sys
from pathlib import PurePath
from isal import igzip
import bionumpy as bnp
import numpy as np
#import numpy_indexed as npi
import resource
import logging

from graph_kmer_index import CollisionFreeKmerIndex, CounterKmerIndex, KmerIndex
from graph_kmer_index.collision_free_kmer_index import MinimalKmerIndex
from graph_kmer_index.index_bundle import IndexBundle
from shared_memory_wrapper import from_file


def read_fasta(file_name):
    i = 0
    with open(file_name, "rb") as f:
        for line in f:
            if line[0] != 62:
                i += 1
                yield line



def remap_array(array, from_values, to_values):
    index = np.digitize(array.ravel(), from_values, right=True)
    return to_values[index].reshape(array.shape)



def log_memory_usage_now(logplace=""):
    memory = int(resource.getrusage(resource.RUSAGE_SELF).ru_maxrss) / 1000000
    logging.info("Memory usage (%s): %.4f GB" % (logplace, memory))


def _get_kmer_index_from_args(args):
    # allowing multiple ways to specify kmer index, try to get the index
    if isinstance(args.kmer_index, KmerIndex):
        kmer_index = args.kmer_index
        kmer_index.convert_to_int32()
        kmer_index.remove_ref_offsets()
        return kmer_index

    if args.kmer_index is None:
        if args.index_bundle is None:
            logging.error("Either a kmer index (-i) or an index bundle (-b) needs to be specified")
            sys.exit(1)
        else:
            kmer_index = IndexBundle.from_file(args.index_bundle).indexes["kmer_index"]
            kmer_index.convert_to_int32()
            kmer_index.remove_ref_offsets()  # not needed, will save us some memory
    else:

        cls = KmerIndex
        if "minimal" in args.kmer_index:
            cls = MinimalKmerIndex
        try:
            kmer_index = cls.from_file(args.kmer_index)
            kmer_index.convert_to_int32()
            kmer_index.remove_ref_offsets()  # not needed, will save us some memory
        except:
            kmer_index = from_file(args.kmer_index)
            assert isinstance(kmer_index, CounterKmerIndex)
            logging.info("Kmer index is counter index")

    return kmer_index


def get_kmer_hashes_from_chunk_sequence(chunk_sequence, kmer_size):
    hashes = bnp.sequence.get_kmers(
        bnp.as_encoded_array(chunk_sequence, bnp.DNAEncoding), kmer_size).ravel().raw().astype(np.uint64)
    logging.debug("N hashes: %d" % len(hashes))
    return hashes


def open_file(filename):
    path = PurePath(filename)
    suffix = path.suffixes[-1]

    if suffix == ".gz":
        suffix = path.suffixes[-2]

    try:
        buffer_type = bnp.io.files._get_buffer_type(suffix)
        logging.info("Using buffer type %s" % buffer_type)
    except RuntimeError:
        logging.error("Unsupported file suffix %s" % suffix)
        raise

    if suffix == ".fa":
        # override buffer type for some performance gain
        buffer_type = bnp.TwoLineFastaBuffer
        logging.info("Using buffer type TwoLineFastaBuffer")

    logging.info("Using igzip")
    open_func = igzip.open if path.suffixes[-1] == ".gz" else open
    reader = bnp.io.parser.NumpyFileReader(open_func(filename, "rb"), buffer_type)
    reader.set_prepend_mode()  # important for performance
    return reader



import logging
import sys
logging.basicConfig(stream=sys.stdout, level=logging.INFO, format='%(asctime)s %(levelname)s: %(message)s')


import os
import time
import tqdm
from .util import _get_kmer_index_from_args, get_kmer_hashes_from_chunk_sequence, open_file, log_memory_usage_now
import numpy as np
import warnings
warnings.filterwarnings("ignore", category=np.VisibleDeprecationWarning) 

import pyximport
pyximport.install()

from graph_kmer_index.collision_free_kmer_index import CounterKmerIndex
import argparse
from graph_kmer_index import KmerIndex
from shared_memory_wrapper.shared_memory import remove_shared_memory_in_session, remove_shared_memory
from shared_memory_wrapper.shared_memory import get_shared_pool
from shared_memory_wrapper import object_to_shared_memory, object_from_shared_memory
import bionumpy as bnp
from kmer_mapper.mapper import map_kmers_to_graph_index
from shared_memory_wrapper.shared_array_map_reduce import additative_shared_array_map_reduce


def main():
    run_argument_parser(sys.argv[1:])


def map_cpu(args, kmer_index, chunk_sequence_name):
    """Maps a sequence stored in shared memory to the kmer index"""
    kmer_size = args["kmer_size"]
    logging.debug("Starting _mapper with chunk %s" % chunk_sequence_name)
    t = time.perf_counter()
    chunk_sequence = object_from_shared_memory(chunk_sequence_name).get_data().sequence
    logging.debug("N sequences in chunk: %d" % len(chunk_sequence))
    # Replace N's with A to "allow" reads with N. Assumes there are very few N in reads (which is usually true)
    chunk_sequence[chunk_sequence == "N"] = "A"
    hashes = get_kmer_hashes_from_chunk_sequence(chunk_sequence, kmer_size)
    logging.debug("Removing %s" % chunk_sequence_name)
    remove_shared_memory(chunk_sequence_name)  # removev now to save some memory usage

    t0 = time.perf_counter()
    if isinstance(kmer_index, CounterKmerIndex):
        kmer_index.counter.count(hashes)
        mapped = kmer_index.counter._values
        logging.debug("Mapped with counter. Got values of length %d" % len(mapped))
    else:
        mapped = map_kmers_to_graph_index(kmer_index, kmer_index.max_node_id(), hashes)

    logging.debug("Mapping %d hashes took %.3f sec" % (len(hashes), time.perf_counter()-t0))

    logging.debug("Chunk of %d reads took %.2f sec" % (len(chunk_sequence), time.perf_counter()-t))
    return mapped


def map_gpu(index, chunks, k, hash_map_size, map_reverse_complements=False):
    """Maps a generator of chunks to the index (no shared memory)"""
    from .gpu_counter import GpuCounter
    logging.info("Making counter")
    counter = GpuCounter.from_kmers_and_nodes(index._kmers, index._nodes, k)
    counter.initialize_cuda(hash_map_size)
    logging.info("CUDA counter initialized")

    t_start = time.perf_counter()
    for i, chunk in enumerate(chunks):
        t0 = time.perf_counter()
        hashes = get_kmer_hashes_from_chunk_sequence(chunk.get_data().sequence, k)
        logging.debug("Time to get hashes for chunk: %.5f", (time.perf_counter()-t0))
        t1 = time.perf_counter()
        counter.count(hashes, count_revcomps=map_reverse_complements)
        logging.debug("Time to count %d hashes for chunk: %.10f" % (len(hashes), time.perf_counter()-t1))
        logging.debug("GPU: Whole chunk finished in %.5f sec", (time.perf_counter()-t0))

    logging.info("Time spent only on hashing and counting hashes: %.5f" % (time.perf_counter()-t_start))
    return counter.get_node_counts(min_nodes=index.max_node_id())


def map_bnp(args):
    if args.debug:
        logging.info("Will print debug log")
        logging.getLogger().setLevel(logging.DEBUG)

    k = args.kmer_size
    start_time = time.perf_counter()
    kmer_index = _get_kmer_index_from_args(args)

    n_bytes = os.stat(args.reads).st_size
    if args.reads.endswith(".gz"):
        n_bytes *= 6.5  # rough estimate for gzipped to give a progress
    approx_number_of_chunks = int(n_bytes / args.chunk_size)
    logging.info("N bytes of reads: %d" % n_bytes)
    logging.info("Approx number of chunks of %d bytes: %d" % (args.chunk_size, approx_number_of_chunks))

    if args.gpu:
        import cupy as cp
        bnp.set_backend(cp)
        file = open_file(args.reads)
        chunks = file.read_chunks(min_chunk_size=args.chunk_size)
        node_counts = map_gpu(kmer_index, chunks, k, args.gpu_hash_map_size, args.map_reverse_complements)
    else:
        assert not args.map_reverse_complements, "Mapping reverse complements only supported with GPU-mode for now"
        file = open_file(args.reads)
        chunks = (object_to_shared_memory(raw_chunk) for
                  raw_chunk in file.read_chunks(min_chunk_size=args.chunk_size))
        chunks = tqdm.tqdm(chunks, total=approx_number_of_chunks)

        if isinstance(kmer_index, KmerIndex):
            initial_data = np.zeros(kmer_index.max_node_id()+1, dtype=np.uint32)
        else:
            initial_data = np.zeros_like(kmer_index.counter._values)

        args_dict = vars(args)
        args_dict.pop("func")
        t_before_map = time.perf_counter()
        node_counts = additative_shared_array_map_reduce(map_cpu,
                                                         chunks,
                                                         initial_data,
                                                         (args_dict, kmer_index),
                                                         n_threads=args.n_threads,
                                                         queue_size_factor=2.0
                                                         )
        logging.info("Time spent only on hashing and counting hashes: %.4f" % (time.perf_counter()-t_before_map))

        if isinstance(kmer_index, CounterKmerIndex):
            # node counts is not node counts, but kmer counts, get node counts
            t = time.perf_counter()
            kmer_index.counter._values = node_counts
            node_counts = kmer_index.get_node_counts()
            logging.info("Time spent getting node counts in the end: %.3f" % (time.perf_counter()-t))

    if args.output_file is None:
        return node_counts

    np.save(args.output_file, node_counts)
    logging.info("Saved node counts to %s.npy" % args.output_file)
    logging.info("Spent %.3f sec in total mapping kmers using %d threads" % (time.perf_counter()-start_time, args.n_threads))
    #remove_shared_memory_in_session()  # not necessary, all memory is freed manually


def run_argument_parser(args):

    parser = argparse.ArgumentParser(
        description='Kmer Mapper',
        prog='kmer_mapper',
        formatter_class=lambda prog: argparse.HelpFormatter(prog, max_help_position=50, width=100))

    subparsers = parser.add_subparsers()
    subparser = subparsers.add_parser("map", help="Map reads to a kmer index")
    subparser.add_argument("-i", "--kmer-index", required=False)
    subparser.add_argument("-b", "--index-bundle", required=False)
    subparser.add_argument("-f", "--reads", required=True, help="Reads in .fa, .fq, .fa.gz, or fq.gz format")
    subparser.add_argument("-k", "--kmer-size", required=False, default=31, type=int)
    subparser.add_argument("-t", "--n-threads", required=False, default=16, type=int)
    subparser.add_argument("-c", "--chunk-size", required=False, type=int, default=2500000,
                           help="N bytes to process in each chunk")
    subparser.add_argument("-o", "--output-file", required=True)
    subparser.add_argument("-d", "--debug", required=False, help="Set to True to print debug log")
    subparser.add_argument("-I", "--max-hits-per-kmer", required=False, default=1000, type=int,
                           help="Ignore kmers that have more than this amount of hits in index")
    subparser.add_argument("-g", "--gpu", default=False, type=bool,
                           help="Set to True to use GPU-counting. Experimental."
                           " Requires suitable hardware and dependencies.")
    subparser.add_argument("-s", "--gpu-hash-map-size", default=0, type=int, help="Can be overriden to set GPU hash map size. "
                           "Set to a lower number to decrease GPU memory requirements. Higher number makes things faster")
    subparser.add_argument("-r", "--map-reverse-complements", default=False, type=bool,
                            help="Also count kmers in reverse complement of reads. "
                                 "Default False. Not necessary if index contains reverse complements.")
    subparser.set_defaults(func=map_bnp)

    if len(args) == 0:
        parser.print_help()
        sys.exit(1)

    args = parser.parse_args(args)
    args.func(args)

    remove_shared_memory_in_session()



import time

import numpy as np
from bionumpy.sequence import get_kmers
from bionumpy.encodings.alphabet_encoding import ACTGEncoding
from bionumpy.encodings._legacy_encodings import ACTGTwoBitEncoding
import bionumpy as bnp
from kmer_mapper.util import get_kmer_hashes_from_chunk_sequence


def _test2():

    numeric_sequence = np.array(np.arange(35) % 4, dtype=np.uint8)
    #numeric_sequence = np.array([1, 1, 1, 1], dtype=np.uint8)
    k = 31
    hash = get_kmers(numeric_sequence, k).ravel()
    print(hash)

    hashes_complement = ACTGTwoBitEncoding.complement(hash) & np.uint64(4**k-1)
    print("----")
    print(hashes_complement)

    reverse_complement = ((numeric_sequence+2)%4)[::-1]
    hash_convolve = np.convolve(reverse_complement.astype(np.uint64), 4**np.arange(k).astype(np.uint64), mode='valid')
    print("----")
    print(hash_convolve[::-1])



def new():
    k = 31
    import bionumpy as bnp
    #file = bnp.open("test.fa")
    file = bnp.open("test.fa")
    for chunk in file.read_chunks(min_chunk_size=150000000):

        hash = get_kmer_hashes_from_chunk_sequence(chunk.sequence, k)
        #hash = bnp.sequence.get_kmers(bnp.as_encoded_array(chunk.sequence, ACTGEncoding), k).ravel().raw()
        # needed for same result
        #hash = ACTGTwoBitEncoding.complement(hash) & np.uint64(4 ** k - 1)
        print(hash)
        return hash


def test():
    hashes = []
    for func in [new]:
        t0 = time.perf_counter()
        result = func()
        hashes.append(result)
        print(func, time.perf_counter()-t0)





if __name__ == "__main__":
    test()


from kmer_mapper.mapper import map_kmers_to_graph_index
import pytest
from graph_kmer_index import sequence_to_kmer_hash, letter_sequence_to_numeric
from graph_kmer_index.read_kmers import ReadKmers
from Bio import Seq
import numpy as np
import time
from graph_kmer_index import KmerIndex
from graph_kmer_index import FlatKmers

def make_dummy_fasta():
    sequences = ["AAAAATTCCACAccagAT", "AACCCAcaAAAACCCACA", "ACTACACACAACCATT", "ACACCATGGAgAGGAATTAC"]
    sequences = ["ACTATCAGCATCAGCATCAGCAGCATGCTACGACGACTACGACTACGGACGACT"]
    #sequences = ["AAAAC"] # , "AACCC", "ACT"]
    file_name = "testfasta.fa.tmp"
    f = open(file_name, "w")
    for i, seq in enumerate(sequences):
        f.write(">seq" + str(i) + "\n" + seq + "\n")

    f.close()

    return sequences, file_name


def true_kmers_from_sequences(file_name, k=3):
    kmers = ReadKmers.from_fasta_file(file_name, k=k)
    kmers = np.concatenate(list(kmers))
    return np.unique(kmers, return_counts=True)


@pytest.mark.skip()
def test_map_to_kmer_index():
    node_kmers = ["ACT", "CTT", "cCG", "ATT"]
    nodes = np.arange(len(node_kmers), dtype=np.uint32)
    node_kmers = np.array([sequence_to_kmer_hash(kmer) for kmer in node_kmers], dtype=np.uint64)
    flat_kmers = FlatKmers(node_kmers, nodes, ref_offsets=np.arange(len(nodes), dtype=np.uint64))
    kmer_index = KmerIndex.from_flat_kmers(flat_kmers, modulo=21)
    kmer_index.convert_to_int32()

    assert kmer_index.get(sequence_to_kmer_hash("ccg"))[0][0] == 2

    print("Node kmers", node_kmers)
    node_counts = map_kmers_to_graph_index(kmer_index, 100, node_kmers, 1000)
    print(node_counts)


#test_get_kmers_from_fasta()
#test_get_kmers_from_fasta()
#test_map_to_index()
#test_map_to_kmer_index()






from setuptools import setup
from distutils.core import setup, Extension
from Cython.Build import cythonize
from distutils.extension import Extension
from Cython.Distutils import build_ext
import numpy as np


ext_modules=[
    Extension("kmer_mapper.mapper",
              ["kmer_mapper/mapper.pyx"],
              libraries=["m"],
              extra_compile_args = ["-O3", "-march=native"],
              )
      ]

setup(name='kmer_mapper',
      version='0.0.36',
      description='Kmer Mapper',
      url='http://github.com/ivargr/kmer_mapper',
      author='Ivar Grytten',
      author_email='',
      license='MIT',
      packages=["kmer_mapper"],
      zip_safe=False,
      install_requires=['numpy', 'bionumpy>=0.2.7', 'cython', 'graph_kmer_index>=0.0.22',
                        'shared_memory_wrapper>=0.0.21', 'tqdm', 'isal'],
      classifiers=[
            'Programming Language :: Python :: 3'
      ],
      entry_points={
            'console_scripts': ['kmer_mapper=kmer_mapper.command_line_interface:main']
      },
      cmdclass = {"build_ext": build_ext},
      ext_modules = ext_modules,
      include_dirs=np.get_include(),
      #ext_modules=cythonize("kmer_mapper/mapper.pyx"),
)

"""
rm -rf dist
python3 setup.py sdist bdist_wheel
auditwheel repair --plat manylinux_2_17_x86_64 dist/kmer_mapper-*-cp38-cp38-linux_x86_64.whl
rm dist/*.whl
mv wheelhouse/* dist
python3 -m twine upload --repository pypi dist/*

"""



import numpy as np
import logging


class GpuCounter:
    def __init__(self, unique_kmers, kmers, nodes, k):
        self.unique_kmers = unique_kmers
        self.kmers = kmers
        self.nodes = nodes
        self.counter = None
        self.k = k

    def initialize_cuda(self, modulo):
        from cucounter.counter import Counter
        print("N unique kmers: %d" % len(self.unique_kmers))
        self.counter = Counter(self.unique_kmers, modulo)

    @classmethod
    def from_kmers_and_nodes(cls, kmers, nodes, k) -> "GpuCounter":
        unique_kmers = np.unique(kmers)
        return cls(unique_kmers, kmers, nodes, k)

    def count(self, kmers, count_revcomps=False):
        self.counter.count(kmers, count_revcomps, self.k)

    def get_node_counts(self, min_nodes=0):
        # get counts in chunks to lower memory
        counts = np.zeros(len(self.kmers), dtype=np.uint32)
        chunk_size = 10_000_000
        start = 0
        for chunk in np.array_split(self.kmers, max(1, len(self.kmers)//chunk_size)):
            print("Querying chunk %d-%d" % (start, start+len(chunk)))
            counts[start:start+len(chunk)] = self.counter[chunk]
            start += len(chunk)

        logging.info("Doing bincount")
        return np.bincount(self.nodes, counts, minlength=min_nodes)

